import json
import sys
import time

import pydash

import constants
from hud import Hud

from pygame import mixer

from levels import Level
from menu import MainMenu, LevelSelection
from logo import Logo
from pygame.locals import *
from player import Player
import sprites
from utils import *
from wallet import Wallet

from pygame_vkeyboard import *


class Game():
    def __init__(self):
        pygame.init()
        pygame.display.set_caption("Indiana Bones")

        # initalize both menu screens
        self.main_menu = MainMenu(self)
        self.level_select = LevelSelection(self)

        # set the current menu to main menu as the game starts there
        self.curr_menu = self.main_menu

        # parameters needed for state destinction
        # the game is running
        self.running = True
        # the game is in a menu
        self.inMenu = True
        # the user is actively playing in a level
        self.playing = False
        # the game is paused
        self.paused = False
        # the player wasted all their lives and the game is over
        self.game_over = False
        # the player won a level and goes into the success screen
        self.success = False
        # the game shows the controls from the pause menu
        self.show_controls = False
        # bool used for the stopwatch, indicates whether or not a level has been paused before.
        self.prev_paused = False
        # indicates to the stopwatch that the player got revived
        self.revived = False
        # the starttime when the gameloop starts
        self.starttime = pygame.time.get_ticks()

        # initializing of the HUD which is displayed in-game
        self.hud = Hud()

        #the name the player types into the highscore board
        self.highscore_name = ""

        # intializing the coin wallet that handles the players coin count
        self.wallet = Wallet()

        # initialization of game and GUI Sounds and Sound Channels
        self.select_sound = pygame.mixer.Sound("assets/sounds/select.wav")
        self.enter_sound = pygame.mixer.Sound("assets/sounds/validate.wav")

        self.leave_sound = pygame.mixer.Sound("assets/sounds/leave_game.wav")
        self.crouch_sound = pygame.mixer.Sound("assets/sounds/crouch.wav")
        self.crouch_sound.set_volume(0.3)

        self.game_over_sound = pygame.mixer.Sound("assets/sounds/gameover.wav")
        self.game_over_sound.set_volume(0.3)

        self.success_sound = pygame.mixer.Sound("assets/sounds/level_success.wav")
        self.success_sound.set_volume(0.5)

        self.fire_sound = pygame.mixer.Sound("assets/sounds/fire.wav")
        self.fire_sound.set_volume(0.4)
        self.fire_channel = pygame.mixer.Channel(6)

        self.character_change_sound = pygame.mixer.Sound("assets/sounds/char_change.wav")
        self.character_change_sound.set_volume(0.7)

        pygame.mixer.set_num_channels(8)

        # initialization of the clock handled for counting the time needed for a level
        self.clock = pygame.time.Clock()

        # timer needed for joystick debouncing in menu navigation
        self.debouncing_timer = 0

        # initialize the pop up window used to display controls
        self.init_pop_up_window()

        # set input buttons appropriate for selected input method in constants.py
        if constants.INPUT_METHOD == "KEYBOARD":
            self.PAUSE_KEY = pygame.K_ESCAPE
            self.RETURN_KEY = pygame.K_RETURN
            self.UP_KEY = pygame.K_UP
            self.DOWN_KEY = pygame.K_DOWN
            self.CHANGE_PLAYER_KEY = pygame.K_c
            self.COIN_KEY = pygame.K_n
        elif constants.INPUT_METHOD == "JOYSTICK":
            self.joystick_player_1 = pygame.joystick.Joystick(0)
            # The program crashes if this is uncommented and there is only one joystick
            self.joystick_player_2 = pygame.joystick.Joystick(1)

            # Comment this if there is only one joystick, s.t. the game doesn't crash. This is only sensible when testing the game
            # self.joystick_player_2 = pygame.joystick.Joystick(0)
            self.PAUSE_KEY = 4
            self.RETURN_KEY = 0  # This is the black key on the Fab-Arcade, lets see if this is sensible
            self.CHANGE_PLAYER_KEY = 2  # This is the blue key on Fab-Arcade
            self.COIN_KEY = 5
            # Attention, these are not Buttons but AXES!
            self.UP_KEY = 8
            self.DOWN_KEY = 8

    # Method that initializes the pop up window used to display game controls
    def init_pop_up_window(self):
        self.pop_up_window = pygame.image.load("assets/hud/pop_up_window.png")
        self.pop_up_window = pygame.transform.scale(self.pop_up_window,
                                                    (self.pop_up_window.get_width() * 1.5,
                                                     self.pop_up_window.get_height() * 1.5))

        controls_overview = pygame.image.load("assets/hud/controls_hud.png")
        controls_overview = pygame.transform.scale(controls_overview, (
            controls_overview.get_width() / 2, controls_overview.get_height() / 2))

        jump_surf = create_text_surf("JUMP", 13, (255, 255, 255))
        special1_surf = create_text_surf("DRAG BOX", 13, (255, 255, 255))
        special2_surf = create_text_surf("/", 13, (255, 255, 255))
        special3_surf = create_text_surf("CROUCH", 13, (255, 255, 255))
        switch_surf1 = create_text_surf("SWITCH CH.", 13, (255, 255, 255))
        switch_surf2 = create_text_surf("(SP)", 13, (255, 255, 255))
        enter_surf = create_text_surf("ENTER", 13, (255, 255, 255))
        pause_control_surf = create_text_surf("PAUSE GAME", 13, (255, 255, 255))
        left_surf = create_text_surf("L", 13, (255, 255, 255))
        right_surf = create_text_surf("R", 13, (255, 255, 255))
        climb_up_surf = create_text_surf("CLIMB UP", 13, (255, 255, 255))
        climb_down_surf = create_text_surf("CLIMB DOWN", 13, (255, 255, 255))
        note_surf = create_text_surf("Press ENTER to close", 13, (237, 189, 78))

        self.pop_up_window.blit(controls_overview, (190, 70))
        self.pop_up_window.blit(jump_surf, (485, 90))
        self.pop_up_window.blit(special1_surf, (480, 270))
        self.pop_up_window.blit(special2_surf, (510, 290))
        self.pop_up_window.blit(special3_surf, (490, 310))
        self.pop_up_window.blit(switch_surf1, (380, 110))
        self.pop_up_window.blit(switch_surf2, (405, 90))
        self.pop_up_window.blit(enter_surf, (400, 280))
        self.pop_up_window.blit(pause_control_surf, (300, 50))
        self.pop_up_window.blit(left_surf, (169, 155))
        self.pop_up_window.blit(right_surf, (335, 155))
        self.pop_up_window.blit(climb_up_surf, (213, 70))
        self.pop_up_window.blit(climb_down_surf, (205, 285))
        self.pop_up_window.blit(note_surf, (30, self.pop_up_window.get_height() - 40))

    # we handle the sprites in a global python file with Sprite Groups and Arrays
    # These have to be reset after selecting/restarting a new level
    def reset_sprites_globals(self):
        sprites.all_sprites.empty()
        sprites.obstacles.empty()
        sprites.players.empty()
        sprites.buttons.empty()
        sprites.ladders.empty()
        sprites.doors.empty()
        sprites.finishes.empty()
        sprites.coins.empty()
        sprites.dead.empty()
        sprites.fires.empty()
        sprites.slideables.empty()
        sprites.halfTiles.empty()

        del sprites.starts[:]
        del sprites.background[:]
        del sprites.decorations[:]

        sprites.logo.empty()

    def select_level(self, level_tuple):
        # level tuple contains path to the .tmx file of the level that was created using Tiled
        # and to a door_data json that contains the logical connections of the level, i.e. which buttons
        # opens which doors
        self.wallet.decrease_balance()
        self.reset_sprites_globals()

        # self.level_tuple is to use when reinitializing after game over
        self.level_tuple = level_tuple

        level = Level(*level_tuple)
        level.initialize_level()

        # the background surface is only rendered once and then blitted onto the display each frame
        # it contains all static elements of the level, i.e. the background, ladders, starts and finishes,
        # normal tiles to move on (obstacles), static dead tiles (spikes) and decorations
        background_surface = pygame.Surface((WIDTH, HEIGHT))
        for background_tile in sprites.background:
            background_surface.blit(background_tile.surf, background_tile.pos)
        for ladder_tile in sprites.ladders:
            background_surface.blit(ladder_tile.surf, ladder_tile.rect)
        for finish in sprites.finishes:
            background_surface.blit(finish.surf, finish.rect)
        for start in sprites.starts:
            background_surface.blit(start.surf, dest=start.pos)
        for obstacle in sprites.obstacles:
            if not sprites.slideables.has(obstacle):
                background_surface.blit(obstacle.surf, obstacle.rect)
        for dead_tile in sprites.dead:
            if not dead_tile.animation_frames:
                background_surface.blit(dead_tile.surf, dead_tile.rect.topleft - vec(
                    (dead_tile.surf.get_width() - dead_tile.rect.width) // 2,
                    dead_tile.surf.get_height() - dead_tile.rect.height))
        for decoration in sprites.decorations:
            background_surface.blit(decoration.surf, dest=decoration.pos)

        # this main surface is blitted onto the display each frame first
        self.main_surface = background_surface

        game_mode = self.main_menu.state
        if INPUT_METHOD == "JOYSTICK":
            self.P1 = Player(player_1=True, start_position=sprites.starts[0].pos, game_mode=game_mode,
                             joystick=self.joystick_player_1)
            self.P2 = Player(player_2=True, start_position=sprites.starts[0].pos, game_mode=game_mode,
                             joystick=self.joystick_player_2)
        else:
            self.P1 = Player(player_1=True, start_position=sprites.starts[0].pos, game_mode=game_mode)
            self.P2 = Player(player_2=True, start_position=sprites.starts[0].pos, game_mode=game_mode)

        # these variables are used to check if the HUD needs to be updated (i.e. when the actual values in the player
        # classes differ from the displayed values in the HUD)
        self.p1_lives_check = self.P1.lives
        self.p2_lives_check = self.P2.lives
        self.p1_coin_check = self.P1.score
        self.p2_coin_check = self.P2.score
        self.p1_bone_check = self.P1.bone_score
        self.p2_bone_check = self.P2.bone_score
        self.hud.hud_init(self.P1.lives, self.P2.lives, self.P1.score + self.P2.score,
                          self.P1.bone_score + self.P2.bone_score)

        sprites.all_sprites.add(self.P1, self.P2)
        sprites.players.add(self.P1, self.P2)

    # the loop that runs when the game is paused
    def pause_loop(self):

        #saving the time at which the game enters the pause loop, so we can calculate the time passed in the pause menu
        self.pause_start = time.time()

        #cursor appears with an offset to the right of the coordinates of menu points
        cursor_offset_x = -30

        #coordinates to which cursor moves
        resume_x, resume_y = 120, 180
        controls_x, controls_y = 120, 210
        back_to_menu_x, back_to_menu_y = 120, 240
        quit_game_x, quit_game_y = 120, 270

        #the state of the pause menu is on the option resume
        pause_state = "Resume"
        cursor_pos = (resume_x + cursor_offset_x, resume_y)

        # the pause menu
        pause_background = pygame.image.load("assets/hud/pause background.png")
        pause_background = pygame.transform.scale(pause_background, (294, 384))
        pause_surf = create_text_surf("PAUSED", 45, (237, 189, 78))
        pause_background.blit(pause_surf, dest=(42, 40))

        # blurring the current frame of the game
        s = pygame.Surface((1000, 750))  # the size of your rect
        s.set_alpha(128)  # alpha level
        s.fill((0, 0, 0))  # this fills the entire surface
        displaysurface.blit(s, (0, 0))  # (0,0) are the top-left coordinates

        # capturing the last frame of the game before paused to blit so players see the paused game
        pause_snapshot = pygame.Surface((WIDTH, HEIGHT))  # Create image surface
        pause_snapshot.blit(displaysurface, (0, 0))  # Blit portion of the display to the image

        # menu points texts
        resume_surf = create_text_surf("RESUME", 15, (237, 189, 78))
        controls_surf = create_text_surf("CONTROLS", 15, (237, 189, 78))
        main_menu_surf = create_text_surf("GO TO MAIN MENU", 15, (237, 189, 78))
        quit_game_surf = create_text_surf("QUIT GAME", 15, (237, 189, 78))
        cursor_surf = create_text_surf('x', 15, (237, 189, 78))

        # necessary parts for control overview
        pop_up_counter = 0

        while self.paused:
            self.debouncing_timer += 1
            for event in pygame.event.get():
                # this handles menu navigation
                if event.type == QUIT:
                    pygame.quit()
                    sys.exit()
                if INPUT_METHOD == "KEYBOARD":
                    if event.type == pygame.KEYDOWN:
                        if event.key == self.PAUSE_KEY:
                            #pressing the pause key while paused will resume game
                            self.paused = False
                            self.playing = True
                            self.show_controls = False
                        if event.key == self.RETURN_KEY:
                            if pause_state == 'Resume':
                                #resume the game
                                self.enter_sound.play()
                                self.paused = False
                                self.playing = True
                            elif pause_state == 'Back to Main Menu':
                                #leave the game and go to main menu
                                self.leave_sound.play()
                                self.prev_paused = False
                                self.paused = False
                                self.playing = False
                                self.curr_menu = self.main_menu
                                self.inMenu = True
                                # pygame.mixer.stop()
                                mixer.music.unload()
                                mixer.music.load('assets/music/adventure-movie-amp-animation-soundtrack-1230.mp3')
                                mixer.music.set_volume(1.0)
                                mixer.music.play(loops=-1)
                            elif pause_state == "Quit Game":
                                #quit the game entirely
                                self.running, self.playing, self.prev_paused = False, False, False
                                self.paused = False
                                self.run_menu = False
                                self.inMenu = False
                            elif pause_state == "Controls":
                                #display the controls pop up window
                                if pop_up_counter % 2 == 0:
                                    self.show_controls = True
                                    pop_up_counter += 1
                                else:
                                    self.show_controls = False
                                    pop_up_counter += 1
                        #if the pop up window is not open, we can navigate the pause menu
                        if not self.show_controls:
                            if event.key == self.UP_KEY:
                                self.select_sound.play()
                                if pause_state == 'Resume':
                                    cursor_pos = (quit_game_x + cursor_offset_x, quit_game_y)
                                    pause_state = 'Quit Game'
                                elif pause_state == 'Controls':
                                    cursor_pos = (resume_x + cursor_offset_x, resume_y)
                                    pause_state = 'Resume'
                                elif pause_state == 'Back to Main Menu':
                                    cursor_pos = (controls_x + cursor_offset_x, controls_y)
                                    pause_state = 'Controls'
                                elif pause_state == 'Quit Game':
                                    cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                    pause_state = 'Back to Main Menu'
                            if event.key == self.DOWN_KEY:
                                self.select_sound.play()
                                if pause_state == 'Resume':
                                    cursor_pos = (controls_x + cursor_offset_x, controls_y)
                                    pause_state = 'Controls'
                                elif pause_state == 'Controls':
                                    cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                    pause_state = 'Back to Main Menu'
                                elif pause_state == 'Back to Main Menu':
                                    cursor_pos = (quit_game_x + cursor_offset_x, quit_game_y)
                                    pause_state = 'Quit Game'
                                elif pause_state == 'Quit Game':
                                    cursor_pos = (resume_x + cursor_offset_x, resume_y)
                                    pause_state = 'Resume'

                #the same code, but for joystick control
                elif INPUT_METHOD == "JOYSTICK":
                    # the inputs for menu navigation are only processed every 15 frames (otherwise menu handling with
                    # joystick does not work properly as too many inputs are recognized for one action)
                    if self.debouncing_timer > 15:
                        if event.type == pygame.JOYBUTTONDOWN:
                            if event.button == self.PAUSE_KEY:
                                self.debouncing_timer = 0
                                self.paused = False
                                self.playing = True
                                self.show_controls = False
                            if event.button == self.RETURN_KEY:
                                if pause_state == 'Resume':
                                    self.debouncing_timer = 0
                                    self.enter_sound.play()
                                    self.paused = False
                                    self.playing = True
                                elif pause_state == 'Back to Main Menu':
                                    self.debouncing_timer = 0
                                    self.prev_paused = False
                                    self.leave_sound.play()
                                    self.paused = False
                                    self.playing = False
                                    self.curr_menu = self.main_menu
                                    self.inMenu = True
                                    # pygame.mixer.stop()
                                    mixer.music.unload()
                                    mixer.music.load('assets/music/adventure-movie-amp-animation-soundtrack-1230.mp3')
                                    mixer.music.set_volume(1.0)
                                    mixer.music.play(loops=-1)
                                elif pause_state == "Quit Game":
                                    self.prev_paused = False
                                    self.running, self.playing = False, False
                                    self.paused = False
                                    self.run_menu = False
                                    self.inMenu = False
                                elif pause_state == "Controls":
                                    if pop_up_counter % 2 == 0:
                                        self.show_controls = True
                                        pop_up_counter += 1
                                    else:
                                        self.show_controls = False
                                        pop_up_counter += 1
                        elif event.type == pygame.JOYAXISMOTION:
                            if not self.show_controls:
                                # This means DOWN
                                if event.axis == 1 and event.value < -0.95:
                                    self.debouncing_timer = 0
                                    self.select_sound.play()
                                    if pause_state == 'Resume':
                                        cursor_pos = (quit_game_x + cursor_offset_x, quit_game_y)
                                        pause_state = 'Quit Game'
                                    elif pause_state == 'Controls':
                                        cursor_pos = (resume_x + cursor_offset_x, resume_y)
                                        pause_state = 'Resume'
                                    elif pause_state == 'Back to Main Menu':
                                        cursor_pos = (controls_x + cursor_offset_x, controls_y)
                                        pause_state = 'Controls'
                                    elif pause_state == 'Quit Game':
                                        cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                        pause_state = 'Back to Main Menu'
                                # This means UP
                                if event.axis == 1 and event.value > 0.95:
                                    self.debouncing_timer = 0
                                    self.select_sound.play()
                                    self.select_sound.play()
                                    if pause_state == 'Resume':
                                        cursor_pos = (controls_x + cursor_offset_x, controls_y)
                                        pause_state = 'Controls'
                                    elif pause_state == 'Controls':
                                        cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                        pause_state = 'Back to Main Menu'
                                    elif pause_state == 'Back to Main Menu':
                                        cursor_pos = (quit_game_x + cursor_offset_x, quit_game_y)
                                        pause_state = 'Quit Game'
                                    elif pause_state == 'Quit Game':
                                        cursor_pos = (resume_x + cursor_offset_x, resume_y)
                                        pause_state = 'Resume'

            # the pause menu is blitted onto the main display
            displaysurface.blit(pause_snapshot, dest=(0, 0))
            displaysurface.blit(pause_background, (50,30))

            displaysurface.blit(resume_surf, dest=(resume_x, resume_y))
            displaysurface.blit(controls_surf, dest=(controls_x, controls_y))
            displaysurface.blit(main_menu_surf, dest=(back_to_menu_x, back_to_menu_y))
            displaysurface.blit(quit_game_surf, dest=(quit_game_x, quit_game_y))

            displaysurface.blit(cursor_surf, dest=cursor_pos)

            if self.show_controls:
                displaysurface.blit(self.pop_up_window, (
                WIDTH / 2 - self.pop_up_window.get_width() / 2, HEIGHT / 2 - self.pop_up_window.get_height() / 2))

            pygame.display.flip()

    # this is the game loop that is executed while a level is played
    def game_loop(self):

        #stopwatch runs over global time
        # to filter out time passed not actually playing, we introduce a case destinction
        if self.revived:
            #if the player just got revived, we need to account for the time spend in the game over menu
            game_over_duration = time.time() - self.game_over_start
            self.starttime += game_over_duration
            self.revived = False
        else:
            #the loop is started not from the game over screen
            if not self.prev_paused:
                #no time was spend in the pause menu yet
                self.starttime = time.time()
                #checktime = starttime
            else:
                #we subtract time spend in pause menu
                pause_duration = time.time() - self.pause_start
                self.starttime += pause_duration


        while self.playing:
            for event in pygame.event.get():
                if event.type == QUIT:
                    pygame.quit()
                    sys.exit()
                if INPUT_METHOD == "KEYBOARD":
                    if event.type == pygame.KEYDOWN:
                        if event.key == pygame.K_ESCAPE:
                            #open the pause menu
                            self.select_sound.play()
                            self.prev_paused = True
                            self.playing = False
                            self.paused = True
                        if event.key == self.COIN_KEY:
                            #user inserted a coin in the arcade
                            self.wallet.increase_balance()

                        # C is the character change key on keyboard
                        if event.key == self.CHANGE_PLAYER_KEY and self.main_menu.state == "Singleplayer":
                            self.character_change_sound.play()
                            self.P1.active_player = not self.P1.active_player
                            self.P2.active_player = not self.P2.active_player
                        # setting the players to jumping or crouching is handled here, as the KEYDOWN events
                        # in the game are processed here
                        if event.key == self.P1.MOVE_UP and self.P1.active_player:
                            self.P1.jump()
                        if event.key == self.P2.MOVE_UP and self.P2.active_player and not self.P2.crouching:
                            self.P2.jump()
                        if event.key == self.P2.CROUCH and self.P2.active_player:
                            self.P2.crouching = not self.P2.crouching
                            if self.P2.crouching:
                                self.crouch_sound.play()
                                # P2's hitbox is halved when crouching
                                self.P2.rect.height = self.P2.rect.height * .5
                                self.P2.rect.midbottom = self.P2.pos
                            else:
                                # testing rect is used to check if P2 can stand up from crouching without a collision (as the
                                # hitbox is set back to its original size when crouching ends)
                                testing_rect = pygame.Rect(self.P2.rect.left, self.P2.rect.top, self.P2.rect.width,
                                                           self.P2.rect.height * 2)
                                if testing_rect.collidelist([half_tile.rect for half_tile in sprites.halfTiles]) == -1:
                                    self.P2.rect.height = self.P2.rect.height * 2
                                    self.P2.rect.midbottom = self.P2.pos
                                else:
                                    self.P2.crouching = True
                    if event.type == pygame.KEYUP:
                        # cancel jump is used to control the height of the jump. Once the JUMP Button is released,
                        # we cancel the jump
                        if event.key == self.P1.MOVE_UP and self.P1.active_player:
                            self.P1.cancel_jump()
                        if event.key == self.P2.MOVE_UP and self.P2.active_player:
                            self.P2.cancel_jump()
                elif INPUT_METHOD == "JOYSTICK":
                    # This is just the code from the "KEYBOARD" input method case, but adjusted for joystick control.
                    # This is very bad practice, but there was no time for optimization (and it works ;) )
                    # Thus refer to documentation above.
                    if event.type == pygame.JOYBUTTONDOWN:
                        if event.button == self.PAUSE_KEY:
                            self.prev_paused = True
                            self.select_sound.play()
                            self.playing = False
                            self.paused = True
                        if event.button == self.CHANGE_PLAYER_KEY and self.main_menu.state == "Singleplayer":
                            self.character_change_sound.play()
                            self.P1.active_player = not self.P1.active_player
                            self.P2.active_player = not self.P2.active_player
                        # in the multiplayer case, both players are active all the time, thus we have to check
                        # if the input comes from the correct joystick (joystick 1 is for player 2)
                        if event.button == 1 and self.P2.active_player and (
                                self.main_menu.state == "Singleplayer" or event.joy == 1):

                            self.P2.crouching = not self.P2.crouching

                            if self.P2.crouching:
                                self.crouch_sound.play()
                                self.P2.rect.height = self.P2.rect.height * .5
                                self.P2.rect.midbottom = self.P2.pos
                            else:
                                testing_rect = pygame.Rect(self.P2.rect.left, self.P2.rect.top, self.P2.rect.width,
                                                           self.P2.rect.height * 2)
                                if testing_rect.collidelist([half_tile.rect for half_tile in sprites.halfTiles]) == -1:
                                    self.P2.rect.height = self.P2.rect.height * 2
                                    self.P2.rect.midbottom = self.P2.pos
                                else:
                                    self.P2.crouching = True
                        # Same here: we have to check in the multiplayer case that the input comes from the correct
                        # joystick
                        if event.button == 3 and self.P1.active_player and (
                                self.main_menu.state == "Singleplayer" or event.joy == 0):
                            self.P1.jump()
                        elif event.button == 3 and self.P2.active_player and (self.main_menu.state == "Singl"
                                                                                                      "eplayer" or event.joy == 1) and not self.P2.crouching:
                            self.P2.jump()

                    if event.type == pygame.JOYBUTTONUP:
                        if event.button == 3 and self.P1.active_player and event.joy == 0:
                            self.P1.cancel_jump()
                        elif event.button == 3 and self.P2.active_player and event.joy == 1:
                            self.P2.cancel_jump()

            # This checks each frame, if the players are game over, i.e. if any of them has 0 lives
            if self.P1.lives == 0 or self.P2.lives == 0:
                #capture the time at which we enter the game over menu
                self.game_over_start = time.time()
                self.game_over_sound.play()

                fadein_screen = pygame.Surface((1000, 750))
                fadein_screen.set_alpha(1)
                fadein_screen.fill((0, 0, 0))

                fadein = True

                # the fade-in is realized by blitting a semi-transparent black surface onto the display for
                # 320 frames
                for i in range(320):
                    displaysurface.blit(fadein_screen, (0, 0))  # (0,0) are the top-left coordinates
                    pygame.display.flip()

                # we end the fade in with a completely black screen
                bg = pygame.Surface((1000, 750))
                bg.fill((0, 0, 0))
                displaysurface.blit(bg, (0, 0))

                pygame.display.flip()

                # this change state leads to executing the game_over loop from the next frame on
                self.game_over = True
                self.playing = False
                break

            # Now we enter the Rendering of the Level itself

            # First we render the main_surface with all static components onto the display that
            # was created in select_level()
            displaysurface.blit(self.main_surface, dest=(0, 0))

            # Now we render all dynamic sprites (doors, fires, buttons, slideables and the players)
            # They are all implemented as subclasses of the pygame Sprites class
            # Thus their main logic is handled in their respective update() functions

            for door in sprites.doors:
                door.update()

                # doors are added and removed from collision detection according to their state
                if not door.open:
                    if not sprites.obstacles.has(door):
                        sprites.obstacles.add(door)
                else:
                    sprites.obstacles.remove(door)

                # if door open_state <=5, then the opening (or closing) animation takes place or the door is closed.
                # Thus the door must be blitted
                if door.open_state <= 5:
                    displaysurface.blit(door.surf, door.rect)

            for fire in sprites.fires:
                fire.update()
                # sprites are added and removed from DEAD Collision detection according to their visibility state
                if fire.is_visible:
                    if not sprites.dead.has():
                        sprites.dead.add(fire)
                    displaysurface.blit(fire.surf, fire.rect)
                else:
                    sprites.dead.remove(fire)

            # Audio Handling for the Fires
            # If any fire is visible, we play the fire sound
            if any([fire.is_visible for fire in sprites.fires]) and not self.fire_channel.get_busy():
                self.fire_channel.play(self.fire_sound, -1)
            elif all([not fire.is_visible for fire in sprites.fires]):
                self.fire_channel.stop()

            # Buttons must be rendered each frame, as they move when a player stands on them
            for button in sprites.buttons:
                displaysurface.blit(button.surf, button.rect)
                button.update()

            for finish in sprites.finishes:
                finish.update()

            for slideable in sprites.slideables:
                slideable.update()
                displaysurface.blit(slideable.surf, slideable.rect)

            for dead_tile in sprites.dead:
                dead_tile.update()
                if dead_tile.animation_frames:
                    # This is needed as the hitbox of the dead tiles is smaller than their actual (visbile) size
                    # This was decided to do after some playtesting, as the game would be too hard otherwise
                    displaysurface.blit(dead_tile.surf, dead_tile.rect.topleft - vec(
                        (dead_tile.surf.get_width() - dead_tile.rect.width) // 2,
                        dead_tile.surf.get_height() - dead_tile.rect.height))
                if DEBUG:
                    # We render the hitboxes of Dead Tiles in Red if the DEBUG Flag is set
                    surf = pygame.Surface(dead_tile.rect.size)
                    surf.fill((255, 0, 0))
                    displaysurface.blit(surf, dead_tile.rect)

            for player in sprites.players:
                player.update()
                if DEBUG:
                    # draw player hitbox for debugging
                    surf = pygame.Surface(player.rect.size)
                    surf.fill((0, 0, 255))
                    displaysurface.blit(surf, player.rect)
                if player.player_1:
                    displaysurface.blit(player.surf,
                                        player.rect.topleft - vec((player.surf.get_width() - player.rect.width) // 2,
                                                                  player.surf.get_height() - player.rect.height))
                elif player.player_2:
                    displaysurface.blit(player.surf,
                                        player.rect.topleft - vec((player.surf.get_width() - player.rect.width) // 2,
                                                                  player.surf.get_height() - player.rect.height))
                if INPUT_METHOD == "KEYBOARD":
                    player.move()
                elif INPUT_METHOD == "JOYSTICK":
                    player.move_with_joystick()

            # coins need to be rendered each frame as they are animated
            for coin in sprites.coins:
                coin.update()
                displaysurface.blit(coin.surf, coin.rect)

            # if the debug Flag is set, we render the platforms the players touch in RED
            if DEBUG:
                for platform in sprites.obstacles:
                    if platform.touched:
                        surf = pygame.Surface((16, 16))
                        surf.fill((255, 0, 0))
                        rect = surf.get_rect(center=platform.rect.center)
                        displaysurface.blit(surf, rect)
                        platform.touched = False
                    platform.update()

            # Finishes are completed if they collide with players
            if all([finish.completed for finish in sprites.finishes]):
                self.P1.running_channel.stop()
                self.P2.running_channel.stop()
                self.success_sound.play()
                time.sleep(1)
                # We enter the success loop
                self.success = True
                self.paused = False
                self.playing = False

            # HUD update only when needed to be resource efficient
            #check if the player lost a life
            if (self.p1_lives_check != self.P1.lives):
                self.p1_lives_check = self.P1.lives
                self.hud.update_p1_lives(self.P1.lives)
            if (self.p2_lives_check != self.P2.lives):
                self.p2_lives_check = self.P2.lives
                self.hud.update_p2_lives(self.P2.lives)

            #check if the player collected a coin
            if (self.p1_coin_check != self.P1.score or self.p2_coin_check != self.P2.score):
                self.p1_coin_check = self.P1.score
                self.p1_coin_check = self.P2.score
                self.hud.update_coin(self.P1.score + self.P2.score)

            #check if the player collected a bone
            if (self.p1_bone_check != self.P1.bone_score or self.p2_bone_check != self.P2.bone_score):
                self.p1_bone_check = self.P1.bone_score
                self.p2_bone_check = self.P2.bone_score
                self.hud.update_bones(self.P1.bone_score + self.P2.bone_score)

            #calculate the time spend actually playing
            #get the current time and filter out the times spend elsewhere
            stopwatch = time.time() - self.starttime
            stopwatch_in_sec = int(stopwatch % 60)
            stopwatch_in_min = int(stopwatch / 60)
            self.stopwatch = '{minutes:02d}:{seconds:02d}'.format(minutes=stopwatch_in_min, seconds=stopwatch_in_sec)

            #inform the HUD that the time spend in game increased
            self.hud.update_stopwatch(self.stopwatch)

            #blit the HUD onto the screen
            self.hud.display_hud(self.main_menu.state, self.P1.active_player, self.P2.active_player)

            fps = FramePerSec.get_fps()
            pygame.display.flip()
            FramePerSec.tick(FPS)


    #method for pygame keyboard
    #saves the currently typed in text to the highscore name variable for later use
    def text_consumer_keyboard(self, text):
        self.highscore_name = text


    #the success loop is executed once the player finished a level
    #displays success menu with option for highscore saving, repeating a level, etc
    def success_loop(self):
        mixer.music.unload()
        self.fire_channel.stop()
        self.P1.running_channel.stop()
        self.P2.running_channel.stop()

        # create background elements
        bg = pygame.image.load("assets/level_selection_background/frame_withbackground.png")
        frame = pygame.image.load("assets/hud/level_background.png")
        frame = pygame.transform.scale(frame, (500, 175))

        # create necessary text
        complete_surf = create_text_surf("completed the level in ", 15, (237, 189, 78))
        success_time_surf = create_text_surf(self.stopwatch, 15, (237, 189, 78))
        with_surf = create_text_surf("with", 15, (237, 189, 78))
        success_score_surf = create_text_surf(str(self.P1.score + self.P2.score) + '/' + '30', 15, (237, 189, 78))
        succes_bone_surf = create_text_surf(str(self.P1.bone_score + self.P2.bone_score), 15, (237, 189, 78))
        main_menu_surf = create_text_surf("Go to Main Menu", 15, (237, 189, 78))
        repeat_surf = create_text_surf("Repeat", 15, (237, 189, 78))
        next_surf = create_text_surf("Next Level", 15, (237, 189, 78))
        highscore_surf = create_text_surf("Save High Score ", 15, (237, 189, 78))
        cursor_surf = create_text_surf('x', 15, (237, 189, 78))

        # state, cursor and coordinates of states
        cursor_offset_x = -20

        back_to_menu_x, back_to_menu_y = WIDTH / 2 - (main_menu_surf.get_size()[0] / 2), 360
        repeat_x, repeat_y = WIDTH / 2 - (repeat_surf.get_size()[0] / 2), 320
        next_x, next_y = WIDTH / 2 - (next_surf.get_size()[0] / 2), 300
        highscore_x, highscore_y = WIDTH / 2 - (highscore_surf.get_size()[0] / 2), 250
        complete_x, complete_y = WIDTH / 2 - (complete_surf.get_size()[0] / 2), 120

        success_state = "Save High Score"
        cursor_pos = (highscore_x + cursor_offset_x, highscore_y)

        # images
        logo = Logo()
        logo.move(WIDTH / 2, 100)
        bone = pygame.image.load("assets/hud/goldenbone.png")
        coin = pygame.image.load("assets/hud/coin_hud.png")

        # create one surface to blit (performance enhancement)
        background_surface = pygame.Surface((WIDTH, HEIGHT))
        background_surface.blit(bg, dest=(0, 0))
        background_surface.blit(frame, (background_surface.get_width() / 2 - frame.get_width() / 2, 228))
        background_surface.blit(complete_surf, dest=(complete_x, complete_y))
        background_surface.blit(success_time_surf, (240, 160))
        background_surface.blit(with_surf, dest=(315, 160))
        background_surface.blit(coin, (380, 165))
        background_surface.blit(success_score_surf, dest=(400, 160))
        background_surface.blit(bone, (490, 165))
        background_surface.blit(succes_bone_surf, dest=(520, 160))
        background_surface.blit(main_menu_surf, dest=(back_to_menu_x, back_to_menu_y))
        background_surface.blit(highscore_surf, dest=(highscore_x, highscore_y))
        background_surface.blit(repeat_surf, dest=(repeat_x, repeat_y))
        background_surface.blit(next_surf, dest=(next_x, next_y))

        # prepare highscore dialog
        database_surf = pygame.Surface((WIDTH, HEIGHT))
        database_surf.fill((25, 25, 25))
        instructions_surf = create_text_surf("Type in your name to save your highscore!", 12, (237, 189, 78))
        database_surf.blit(instructions_surf, dest=(100, 70))
        instructions2_surf = create_text_surf("Confirm by pressing CAPSLOCK on Keyboard / START on the Arcade!", 12,
                                              (237, 189, 78))
        database_surf.blit(instructions2_surf, dest=(100, 100))

        # Initializes and activates vkeyboard
        keyboard_layout = VKeyboardLayout(VKeyboardLayout.QWERTY)
        highscore_keyboard = VKeyboard(database_surf, self.text_consumer_keyboard, keyboard_layout,
                                       joystick_navigation=True, show_text=True)

        while self.success:
            self.debouncing_timer += 1
            for event in pygame.event.get():
                if event.type == QUIT:
                    pygame.quit()
                    sys.exit()
                if INPUT_METHOD == "KEYBOARD":
                    # Menu Navigation
                    if event.type == pygame.KEYDOWN:
                        if event.key == self.COIN_KEY:
                            self.wallet.increase_balance()
                        if event.key == pygame.K_RETURN:
                            if success_state == 'Back to Main Menu':
                                self.leave_sound.play()
                                self.prev_paused = False
                                self.success = False
                                self.playing = False
                                self.curr_menu = self.main_menu
                                self.inMenu = True
                                self.success_sound.stop()
                                mixer.music.unload()
                                mixer.music.load('assets/music/adventure-movie-amp-animation-soundtrack-1230.mp3')
                                mixer.music.set_volume(1.0)
                                mixer.music.play(loops=-1)
                            elif success_state == 'Next Level':
                                if self.wallet.balance > 0:
                                    next_index = self.level_select.level_index + 1
                                    self.level_select.level_index += 1

                                    if next_index < len(self.level_select.levels):
                                        self.success = False
                                        self.prev_paused = False
                                        self.select_level(self.level_select.levels[next_index])
                                        mixer.music.load('assets/music/dream_fire_without_beginning_silence.mp3')
                                        mixer.music.set_volume(1.0)
                                        mixer.music.play(loops=-1)
                                        self.success_sound.stop()
                                        self.playing = True
                                    else:
                                        # If there is no next level, we play an error sound
                                        self.level_select.wallet_error_sound.play()
                                else:
                                    self.level_select.wallet_error_sound.play()

                            elif success_state == 'Repeat':
                                if self.wallet.balance > 0:
                                    self.success = False
                                    self.prev_paused = False
                                    self.select_level(self.level_tuple)
                                    mixer.music.load('assets/music/dream_fire_without_beginning_silence.mp3')
                                    mixer.music.set_volume(1.0)
                                    mixer.music.play(loops=-1)
                                    self.success_sound.stop()
                                    self.playing = True
                                else:
                                    self.level_select.wallet_error_sound.play()

                            elif success_state == "Save High Score":
                                self.highscore = True
                                while self.highscore:
                                    events = pygame.event.get()
                                    # Update internal variables
                                    highscore_keyboard.update(events)
                                    # Draw the keyboard
                                    highscore_keyboard.draw(database_surf)

                                    for event in events:
                                        if INPUT_METHOD == "KEYBOARD":
                                            if event.type == pygame.KEYDOWN:
                                                if event.key == pygame.K_CAPSLOCK:

                                                    highscore_data = {'name': self.highscore_name,
                                                                      'time': self.stopwatch,
                                                                      'coins': self.P1.score + self.P2.score,
                                                                      'bone': self.P1.bone_score + self.P2.bone_score}

                                                    with open(self.level_select.level_highscores[
                                                                  self.level_select.level_index], 'r',
                                                              encoding='u8') as f:
                                                        highscores = json.load(f)

                                                    if not highscore_data in highscores:
                                                        highscores.append(highscore_data)

                                                    # This custom comparator sorts the highscores first by bones,
                                                    # if these are equal then by coins, and lastly by time (if coins
                                                    # are also equal
                                                    def comparator(x, y):
                                                        if x["bone"] > y["bone"]:
                                                            return 1
                                                        elif x["bone"] < y["bone"]:
                                                            return -1
                                                        elif x["bone"] == y["bone"]:
                                                            if x["coins"] > y["coins"]:
                                                                return 1
                                                            elif x["coins"] < y["coins"]:
                                                                return -1
                                                            elif x["coins"] == y["coins"]:
                                                                if x["time"] > y["time"]:
                                                                    return 1
                                                                elif x["time"] < y["time"]:
                                                                    return -1
                                                                else:
                                                                    return 0

                                                    # unfortunately we were not smart enough to implement it completely
                                                    # correctly, thus the comparator sorts them right, but in a reverse
                                                    # order. Fortunately, pygame has this reverse flag...
                                                    highscores = pydash.sort(highscores, comparator, reverse=True)

                                                    highscores = highscores[:10]

                                                    with open(self.level_select.level_highscores[
                                                                  self.level_select.level_index], 'w',
                                                              encoding='u8') as f:
                                                        json.dump(highscores, f)

                                                    self.highscore = False

                                    displaysurface.blit(database_surf, dest=(WIDTH / 2 - database_surf.get_width() / 2,
                                                                             HEIGHT / 2 - database_surf.get_height() / 2))
                                    pygame.display.flip()

                                # virtual keyboard adds some strange event in keyboard input mode, this is a workaround
                                continue

                        if event.key == pygame.K_UP:
                            self.select_sound.play()
                            if success_state == 'Save High Score':
                                cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                success_state = 'Back to Main Menu'
                            elif success_state == 'Next Level':
                                cursor_pos = (highscore_x + cursor_offset_x, highscore_y)
                                success_state = 'Save High Score'
                            elif success_state == 'Repeat':
                                cursor_pos = (next_x + cursor_offset_x, next_y)
                                success_state = 'Next Level'
                            elif success_state == 'Back to Main Menu':
                                cursor_pos = (repeat_x + cursor_offset_x, repeat_y)
                                success_state = 'Repeat'

                        if event.key == pygame.K_DOWN:
                            self.select_sound.play()
                            if success_state == 'Save High Score':
                                cursor_pos = (next_x + cursor_offset_x, next_y)
                                success_state = 'Next Level'
                            elif success_state == 'Next Level':
                                cursor_pos = (repeat_x + cursor_offset_x, repeat_y)
                                success_state = 'Repeat'
                            elif success_state == 'Repeat':
                                cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                success_state = 'Back to Main Menu'
                            elif success_state == 'Back to Main Menu':
                                cursor_pos = (highscore_x + cursor_offset_x, highscore_y)
                                success_state = 'Save High Score'
                elif INPUT_METHOD == "JOYSTICK":
                    # once again, very similar code as above. Refer to above documentation.
                    if self.debouncing_timer > 15:
                        if event.type == pygame.JOYBUTTONDOWN:
                            if event.button == self.COIN_KEY:
                                self.wallet.increase_balance()
                            if event.button == self.RETURN_KEY:
                                self.debouncing_timer = 0
                                if success_state == 'Back to Main Menu':
                                    self.prev_paused = False
                                    self.leave_sound.play()
                                    self.success = False
                                    self.playing = False
                                    self.curr_menu = self.main_menu
                                    self.inMenu = True
                                    self.success_sound.stop()
                                    mixer.music.unload()
                                    mixer.music.load('assets/music/adventure-movie-amp-animation-soundtrack-1230.mp3')
                                    mixer.music.set_volume(1.0)
                                    mixer.music.play(loops=-1)
                                elif success_state == 'Next Level':
                                    if self.wallet.balance > 0:
                                        next_index = self.level_select.level_index + 1
                                        self.level_select.level_index += 1

                                        if next_index < len(self.level_select.levels):
                                            self.prev_paused = False
                                            self.success = False
                                            self.select_level(self.level_select.levels[next_index])
                                            mixer.music.load('assets/music/dream_fire_without_beginning_silence.mp3')
                                            mixer.music.set_volume(1.0)
                                            mixer.music.play(loops=-1)
                                            self.success_sound.stop()
                                            self.playing = True
                                        else:
                                            self.level_select.wallet_error_sound.play()
                                    else:
                                        self.level_select.wallet_error_sound.play()


                                elif success_state == 'Repeat':
                                    if self.wallet.balance > 0:
                                        self.success = False
                                        self.prev_paused = False
                                        self.select_level(self.level_tuple)
                                        mixer.music.load('assets/music/dream_fire_without_beginning_silence.mp3')
                                        mixer.music.set_volume(1.0)
                                        mixer.music.play(loops=-1)
                                        self.success_sound.stop()
                                        self.playing = True
                                    else:
                                        self.level_select.wallet_error_sound.play()

                                elif success_state == "Save High Score":
                                    self.highscore = True
                                    while self.highscore:
                                        events = pygame.event.get()
                                        keyboard_events = []
                                        for event in events:
                                            if INPUT_METHOD == "JOYSTICK":
                                                if event.type == pygame.JOYAXISMOTION:
                                                    # we have to translate the joystick events to button presses, as
                                                    # the virtual keyboard can only handle keyboard button presses
                                                    if event.axis == 1 and event.value < -0.95:
                                                        keyboard_events.append(pygame.event.Event(pygame.KEYDOWN,
                                                                                                  {"key": pygame.K_UP,
                                                                                                   "unicode": "",
                                                                                                   "mod": 0}))

                                                    if event.axis == 1 and event.value > 0.95:
                                                        keyboard_events.append(pygame.event.Event(pygame.KEYDOWN,
                                                                                                  {"key": pygame.K_DOWN,
                                                                                                   "unicode": "",
                                                                                                   "mod": 0}))

                                                    if event.axis == 0 and event.value < -0.95:
                                                        keyboard_events.append(pygame.event.Event(pygame.KEYDOWN,
                                                                                                  {"key": pygame.K_LEFT,
                                                                                                   "unicode": "",
                                                                                                   "mod": 0}))

                                                    if event.axis == 0 and event.value > 0.95:
                                                        keyboard_events.append(pygame.event.Event(pygame.KEYDOWN, {
                                                            "key": pygame.K_RIGHT, "unicode": "", "mod": 0}))

                                                if event.type == pygame.JOYBUTTONDOWN:
                                                    if event.button == self.PAUSE_KEY:

                                                        highscore_data = {'name': self.highscore_name,
                                                                          'time': self.stopwatch,
                                                                          'coins': self.P1.score + self.P2.score,
                                                                          'bone': self.P1.bone_score + self.P2.bone_score}

                                                        with open(self.level_select.level_highscores[
                                                                      self.level_select.level_index], 'r',
                                                                  encoding='u8') as f:
                                                            highscores = json.load(f)

                                                        if not highscore_data in highscores:
                                                            highscores.append(highscore_data)

                                                        def comparator(x, y):
                                                            if x["bone"] > y["bone"]:
                                                                return 1
                                                            elif x["bone"] < y["bone"]:
                                                                return -1
                                                            elif x["bone"] == y["bone"]:
                                                                if x["coins"] > y["coins"]:
                                                                    return 1
                                                                elif x["coins"] < y["coins"]:
                                                                    return -1
                                                                elif x["coins"] == y["coins"]:
                                                                    if x["time"] > y["time"]:
                                                                        return 1
                                                                    elif x["time"] < y["time"]:
                                                                        return -1
                                                                    else:
                                                                        return 0

                                                        highscores = pydash.sort(highscores, comparator, reverse=True)

                                                        highscores = highscores[:10]

                                                        with open(self.level_select.level_highscores[
                                                                      self.level_select.level_index], 'w',
                                                                  encoding='u8') as f:
                                                            json.dump(highscores, f)

                                                        self.highscore = False

                                        # Update internal variables
                                        highscore_keyboard.update(keyboard_events + events)
                                        # Draw the keyboard
                                        highscore_keyboard.draw(database_surf)

                                        displaysurface.blit(database_surf,
                                                            dest=(WIDTH / 2 - database_surf.get_width() / 2,
                                                                  HEIGHT / 2 - database_surf.get_height() / 2))
                                        pygame.display.flip()


                        elif event.type == pygame.JOYAXISMOTION:
                            if event.axis == 1 and event.value < -0.95:
                                self.debouncing_timer = 0
                                self.select_sound.play()
                                if success_state == 'Save High Score':
                                    cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                    success_state = 'Back to Main Menu'
                                elif success_state == 'Next Level':
                                    cursor_pos = (highscore_x + cursor_offset_x, highscore_y)
                                    success_state = 'Save High Score'
                                elif success_state == 'Repeat':
                                    cursor_pos = (next_x + cursor_offset_x, next_y)
                                    success_state = 'Next Level'
                                elif success_state == 'Back to Main Menu':
                                    cursor_pos = (repeat_x + cursor_offset_x, repeat_y)
                                    success_state = 'Repeat'
                            if event.axis == 1 and event.value > 0.95:
                                self.debouncing_timer = 0
                                self.select_sound.play()
                                if success_state == 'Save High Score':
                                    cursor_pos = (next_x + cursor_offset_x, next_y)
                                    success_state = 'Next Level'
                                elif success_state == 'Next Level':
                                    cursor_pos = (repeat_x + cursor_offset_x, repeat_y)
                                    success_state = 'Repeat'
                                elif success_state == 'Repeat':
                                    cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                    success_state = 'Back to Main Menu'
                                elif success_state == 'Back to Main Menu':
                                    cursor_pos = (highscore_x + cursor_offset_x, highscore_y)
                                    success_state = 'Save High Score'

            displaysurface.blit(background_surface, dest=(0, 0))
            logo.update()
            displaysurface.blit(logo.surf, logo.rect)

            displaysurface.blit(cursor_surf, dest=cursor_pos)
            displaysurface.blit(self.wallet.surf, dest=(540, 240))

            pygame.display.flip()
            FramePerSec.tick(FPS)

            # displaysurface.blit(displaysurface, (0, 0))


    #the game over loop is entered once the player lost all of her lives
    def game_over_loop(self):
        # stop all sounds
        mixer.music.unload()
        self.fire_channel.stop()
        self.P1.running_channel.stop()
        self.P2.running_channel.stop()

        #build the menu
        cursor_offset_x = -90
        use_coin_x, use_coin_y = WIDTH / 2, HEIGHT / 2 + 80
        back_to_menu_x, back_to_menu_y = WIDTH / 2, HEIGHT / 2 + 100

        game_over_state = "Use Coin"
        cursor_pos = (use_coin_x + cursor_offset_x, use_coin_y)

        bg_surf = pygame.Surface((WIDTH, HEIGHT))  # the size of your rect
        bg_surf.fill((0, 0, 0))  # this fills the entire surface

        gameover_surf = create_text_surf("GAME OVER", 75, (237, 189, 78))
        revive_surf = create_text_surf("Revive for 1 coin?", 15, (237, 189, 78))
        wallet_surf = create_text_surf(f"Wallet : {self.wallet.balance} coins", 15, (237, 189, 78))
        revive_menu_surf = create_text_surf("Revive!", 15, (237, 189, 78))
        main_menu_surf = create_text_surf("Go to Main Menu", 15, (237, 189, 78))
        cursor_surf = create_text_surf('x', 15, (237, 189, 78))

        bg_surf.blit(gameover_surf, dest=(WIDTH / 2 - gameover_surf.get_width() / 2, 60))
        bg_surf.blit(revive_surf, dest=(WIDTH / 2 - revive_surf.get_width() / 2, 165))
        bg_surf.blit(wallet_surf, dest=(WIDTH / 2 - wallet_surf.get_width() / 2, 185))
        bg_surf.blit(revive_menu_surf, dest=(use_coin_x - revive_menu_surf.get_width() / 2, use_coin_y))
        bg_surf.blit(main_menu_surf, dest=(back_to_menu_x - main_menu_surf.get_width() / 2, back_to_menu_y))

        while self.game_over:
            self.debouncing_timer += 1
            for event in pygame.event.get():
                if event.type == QUIT:
                    pygame.quit()
                    sys.exit()
                if constants.INPUT_METHOD == "KEYBOARD":
                    # Menu Navigation
                    if event.type == pygame.KEYDOWN:
                        if event.key == self.COIN_KEY:
                            self.wallet.increase_balance()
                            bg_surf.fill((0, 0, 0), rect=(
                                WIDTH / 2 - wallet_surf.get_width() / 2, 185, wallet_surf.get_width(),
                                wallet_surf.get_height()))
                            wallet_surf = create_text_surf(f"Wallet : {self.wallet.balance} coins", 15,
                                                           (237, 189, 78))
                            bg_surf.blit(wallet_surf, dest=(WIDTH / 2 - wallet_surf.get_width() / 2, 185))

                        if event.key == pygame.K_RETURN:
                            if game_over_state == 'Use Coin':
                                #use a coin from the wallet to revive the player
                                if self.wallet.balance > 0:
                                    self.game_over = False
                                    self.revived = True
                                    # revive the players
                                    self.P1.lives = 3
                                    self.P2.lives = 3
                                    mixer.music.load('assets/music/dream_fire_without_beginning_silence.mp3')
                                    mixer.music.set_volume(1.0)
                                    mixer.music.play(loops=-1)
                                    self.game_over_sound.stop()
                                    self.playing = True

                            elif game_over_state == 'Back to Main Menu':
                                self.leave_sound.play()
                                self.prev_paused = False
                                self.game_over = False
                                self.playing = False
                                self.curr_menu = self.main_menu
                                self.inMenu = True
                                # pygame.mixer.stop()
                                mixer.music.unload()
                                mixer.music.load('assets/music/adventure-movie-amp-animation-soundtrack-1230.mp3')
                                mixer.music.set_volume(1.0)
                                mixer.music.play(loops=-1)

                        if event.key == pygame.K_UP:
                            self.select_sound.play()
                            if game_over_state == 'Use Coin':
                                cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                game_over_state = 'Back to Main Menu'
                            elif game_over_state == 'Back to Main Menu':
                                cursor_pos = (use_coin_x + cursor_offset_x, use_coin_y)
                                game_over_state = 'Use Coin'
                        if event.key == pygame.K_DOWN:
                            self.select_sound.play()
                            if game_over_state == 'Use Coin':
                                cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                game_over_state = 'Back to Main Menu'
                            elif game_over_state == 'Back to Main Menu':
                                cursor_pos = (use_coin_x + cursor_offset_x, use_coin_y)
                                game_over_state = 'Use Coin'
                #same code but for joystick controls
                elif INPUT_METHOD == "JOYSTICK":
                    # Same as above, basically only Menu Navigation
                    if self.debouncing_timer > 15:
                        if event.type == pygame.JOYBUTTONDOWN:
                            if event.button == self.COIN_KEY:
                                self.wallet.increase_balance()
                                bg_surf.fill((0, 0, 0), rect=(
                                    WIDTH / 2 - wallet_surf.get_width() / 2, 185, wallet_surf.get_width(),
                                    wallet_surf.get_height()))
                                wallet_surf = create_text_surf(f"Wallet : {self.wallet.balance} coins", 15,
                                                               (237, 189, 78))
                                bg_surf.blit(wallet_surf, dest=(WIDTH / 2 - wallet_surf.get_width() / 2, 185))
                            if event.button == self.RETURN_KEY:
                                self.debouncing_timer = 0
                                if game_over_state == 'Use Coin':
                                    #use a coin to revive the player
                                    if self.wallet.balance > 0:
                                        self.game_over = False
                                        self.revived = True
                                        self.P1.lives = 3
                                        self.P2.lives = 3
                                        mixer.music.load('assets/music/dream_fire_without_beginning_silence.mp3')
                                        mixer.music.set_volume(1.0)
                                        mixer.music.play(loops=-1)
                                        self.game_over_sound.stop()
                                        self.playing = True

                                elif game_over_state == 'Back to Main Menu':
                                    self.leave_sound.play()
                                    self.prev_paused = False
                                    self.game_over = False
                                    self.playing = False
                                    self.curr_menu = self.main_menu
                                    self.inMenu = True
                                    # pygame.mixer.stop()
                                    mixer.music.unload()
                                    mixer.music.load('assets/music/adventure-movie-amp-animation-soundtrack-1230.mp3')
                                    mixer.music.set_volume(1.0)
                                    mixer.music.play(loops=-1)
                        elif event.type == pygame.JOYAXISMOTION:
                            if event.axis == 1 and event.value < -.95:
                                self.debouncing_timer = 0
                                self.select_sound.play()
                                if game_over_state == 'Use Coin':
                                    cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                    game_over_state = 'Back to Main Menu'
                                elif game_over_state == 'Back to Main Menu':
                                    cursor_pos = (use_coin_x + cursor_offset_x, use_coin_y)
                                    game_over_state = 'Use Coin'
                            if event.axis == 1 and event.value > .95:
                                self.debouncing_timer = 0
                                self.select_sound.play()
                                if game_over_state == 'Use Coin':
                                    cursor_pos = (back_to_menu_x + cursor_offset_x, back_to_menu_y)
                                    game_over_state = 'Back to Main Menu'
                                elif game_over_state == 'Back to Main Menu':
                                    cursor_pos = (use_coin_x + cursor_offset_x, use_coin_y)
                                    game_over_state = 'Use Coin'

            displaysurface.blit(bg_surf, dest=(0, 0))
            displaysurface.blit(cursor_surf, dest=cursor_pos)
            pygame.display.flip()
            FramePerSec.tick(FPS)

