import random

import pygame
import pytmx
from pytmx import TiledTileLayer
import json

import constants
import sprites
from constants import vec
from player import Player
from sprites import obstacles, all_sprites, background, ladders, buttons, players, finishes, starts, coins, doors, dead, \
    fires, slideables, decorations
import pydash as _


class Level:
    def __init__(self, path_to_level, path_to_door_data):
        # Level needs a Path to the .tmx file created with Tiled and to the Door Data JSON that contains the levels logic
        self.tmx_data = pytmx.load_pygame(path_to_level)
        with open(path_to_door_data, "r") as f:
            door_data = json.load(f)
        # this is an array of all doors in the level (identified by their coordinates) and from which button
        # they are triggered
        self.doors = door_data["doors"]
        # this is an array of all buttons in the level (identified by their coordinates) and which door they do open
        self.buttons = door_data["buttons"]
        self.tile_width = self.tmx_data.tilewidth
        self.tile_height = self.tmx_data.tileheight

    def initialize_level(self):
        door_tiles = []
        # this is done, s.t. we can recognize door tiles while creating a level
        for door in self.doors:
            door_tiles = door_tiles + door["tiles"]

        for layer in self.tmx_data.visible_layers:
            # for documentation of the TmxImporter Methods, refer to https://pytmx.readthedocs.io/en/latest/
            if type(layer) != TiledTileLayer:
                continue
            for x, y, gid in layer:
                tile_surface = self.tmx_data.get_tile_image_by_gid(gid)
                tile_properties = self.tmx_data.get_tile_properties_by_gid(gid)
                if not tile_surface:
                    continue
                if layer.name == "Frontlayer":
                    # Here, for each Tile in the Front Layer, a corresponding Sprite Object is generated that
                    # is put in the correct sprite groups
                    if [x, y] in door_tiles:
                        # with the help of the door data, we find the buttons from which the door can be opened
                        door_info = _.find(self.doors, lambda door: [x, y] in door["tiles"])
                        connected_buttons = _.filter_(self.buttons, lambda x: door_info["id"] in x["connected_door"])
                        tile_sprite = DoorTile(tile_surface, (x * self.tile_width, y * self.tile_height),
                                               door_id=door_info["id"], connected_buttons=connected_buttons)
                        all_sprites.add(tile_sprite)
                        doors.add(tile_sprite)
                    elif tile_properties and tile_properties.get("obstacle"):
                        tile_sprite = ForegroundTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        all_sprites.add(tile_sprite)
                        obstacles.add(tile_sprite)
                    elif tile_properties and tile_properties.get("halfTile"):
                        tile_sprite = ForegroundTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        all_sprites.add(tile_sprite)
                        sprites.halfTiles.add(tile_sprite)
                        obstacles.add(tile_sprite)
                    elif tile_properties and tile_properties.get("ladder"):
                        tile_sprite = ForegroundTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        all_sprites.add(tile_sprite)
                        ladders.add(tile_sprite)
                    elif tile_properties and tile_properties.get("button"):
                        # with the help of the door data we find the door that the button can open
                        button_info = _.find(self.buttons, lambda button: [x, y] in button["tiles"])
                        tile_sprite = ButtonTile(tile_surface, (x * self.tile_width, y * self.tile_height),
                                                 button_info["id"],
                                                 button_info["connected_door"])
                        all_sprites.add(tile_sprite)
                        buttons.add(tile_sprite)
                    elif tile_properties and tile_properties.get("finish"):
                        tile_sprite = FinishTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        all_sprites.add(tile_sprite)
                        finishes.add(tile_sprite)
                    elif tile_properties and tile_properties.get("start"):
                        tile_sprite = StartTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        all_sprites.add(tile_sprite)
                        starts.append(tile_sprite)
                    elif tile_properties and tile_properties.get("collectable"):
                        if tile_properties.get("bone"):
                            tile_sprite = BoneTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        else:
                            tile_sprite = CoinTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        all_sprites.add(tile_sprite)
                        coins.add(tile_sprite)
                    elif tile_properties and tile_properties.get("die"):
                        tile_sprite = DeadTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        all_sprites.add(tile_sprite)
                        dead.add(tile_sprite)
                    elif tile_properties and tile_properties.get("is_fire"):
                        tile_sprite = FireTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        all_sprites.add(tile_sprite)
                        fires.add(tile_sprite)
                    elif tile_properties and tile_properties.get("slideable"):
                        tile_sprite = SlideableTile(tile_surface, (x * self.tile_width, y * self.tile_height))
                        all_sprites.add(tile_sprite)
                        obstacles.add(tile_sprite)
                        slideables.add(tile_sprite)
                    else:
                        # we put everything else that we did not handle into the background (but this should never happen)
                        background.append(BackgroundTile(tile_surface, (x * self.tile_width, y * self.tile_height)))

                    if tile_sprite and tile_properties and tile_properties.get("frames"):
                        # add animations to tiles that have them
                        tile_sprite.animation_frames = [self.tmx_data.get_tile_image_by_gid(frame.gid) for frame in
                                                        tile_properties["frames"]]
                elif layer.name == "Decorations":
                    decorations.append(BackgroundTile(tile_surface, (x * self.tile_width, y * self.tile_height)))
                else:
                    background.append(BackgroundTile(tile_surface, (x * self.tile_width, y * self.tile_height)))


class ForegroundTile(pygame.sprite.Sprite):
    def __init__(self, surface, position):
        super().__init__()
        self.surf = surface
        self.rect = self.surf.get_rect(topleft=position)
        self.touched = False


class SlideableTile(ForegroundTile):
    def __init__(self, surface, position):
        super().__init__(surface, position)
        self.is_wall_right = []
        self.is_wall_left = []

    def update(self):
        # Collision Detection and Box Movement

        player_hits = pygame.sprite.spritecollide(self, players, False)
        obstacle_hits = pygame.sprite.spritecollide(self, obstacles, False)
        obstacle_hits = _.filter_(obstacle_hits, lambda x: not x is self)
        if len(obstacle_hits) > 1:
            self.is_wall_left = [obstacle for
                                 obstacle in
                                 obstacle_hits if
                                 abs(obstacle.rect.top - self.rect.top) <= 1 and obstacle.rect.left < self.rect.left]
            self.is_wall_right = [obstacle
                                  for obstacle in
                                  obstacle_hits if
                                  abs(obstacle.rect.top - self.rect.top) <= 1 and obstacle.rect.left > self.rect.left]
        player_1 = _.find(player_hits, lambda x: x.player_1)

        # this is collision detection with floors (s.t. we do not fall through it)
        if obstacle_hits:
            largest_y_platform = obstacle_hits[0]
            for sprite in obstacle_hits:
                if constants.DEBUG:
                    sprite.touched = True
                if sprite.rect.top > largest_y_platform.rect.top:
                    largest_y_platform = sprite
            if self.rect.top < largest_y_platform.rect.bottom:
                self.rect.bottom = largest_y_platform.rect.top + 1
        else:
            # If there is no floor, we fall (this is basically gravity)
            self.rect.top = self.rect.top + 2

        # if there is a player that we collide with which wants to push (P1.pushing), we are getting pushed
        if player_1 and player_1.pushing and abs(player_1.rect.top - self.rect.top) == 9 and not any(
                [(not box is self) and box.is_pushed for box in sprites.slideables]):
            self.is_pushed = True
        else:
            self.is_pushed = False

        if self.is_pushed:
            # this prevents us from glitching into walls
            if self.is_wall_right:
                self.rect.right = self.is_wall_right[0].rect.left - 1
            elif self.is_wall_left:
                self.rect.left = self.is_wall_left[0].rect.right + 1
            # otherwise we set our position to a position inside of the player s.t. we do not loose collision
            elif player_1.rect.left < self.rect.left:
                self.rect.left = player_1.rect.right - 4
            elif player_1.rect.left > self.rect.left:
                self.rect.right = player_1.rect.left + 4


class FireTile(pygame.sprite.Sprite):
    def __init__(self, surface, position):
        super().__init__()
        self.surf = surface
        self.rect = pygame.Rect(position, (10, 16))

        # if a fire tile is above we have to synchronise the visibility with it
        fire_tile_above = _.find(sprites.fires,
                                 lambda x: (x.rect.topleft == self.rect.topleft - vec(0, 16)) and isinstance(x,
                                                                                                             FireTile))
        if fire_tile_above:
            self.frame_counter = fire_tile_above.frame_counter
        else:
            # otherwise we randomly initialize our visibility state
            self.frame_counter = random.randint(0, 239)
        self.is_visible = False
        self.is_hearable = False
        self.animation_frames = []
        self.cur_animation_index = 0
        self.animation_counter = 0
        self.collected = False
        self.killing = True

    def update(self):
        self.frame_counter += 1
        self.animation_counter += 1

        # every 240 Frames we disappear
        if self.frame_counter == 240:
            self.frame_counter = 0
            self.is_visible = not self.is_visible

        # animation system
        self.cur_animation_index = self.cur_animation_index % len(self.animation_frames)

        if self.animation_counter % 10 == 0:
            self.animation_counter = 0
            self.cur_animation_index = (self.cur_animation_index + 1) % len(self.animation_frames)
        self.surf = self.animation_frames[self.cur_animation_index]


class ButtonTile(pygame.sprite.Sprite):
    def __init__(self, surface, position, button_id, connected_door_id):
        super().__init__()
        self.surf = surface
        self.rect = self.surf.get_rect(topleft=position)
        self.original_x, self.original_y = position
        self.connected_door_id = connected_door_id
        self.pushed = False
        self.cont_push = False
        self.door_open_sound = pygame.mixer.Sound("assets/sounds/door_open.wav")
        self.door_open_sound.set_volume(0.3)
        self.door_close_sound = pygame.mixer.Sound("assets/sounds/door_close.wav")
        self.door_close_sound.set_volume(0.3)
        self.id = button_id

    def update(self):
        player_hits = pygame.sprite.spritecollide(self, players, False)
        slideable_hits = pygame.sprite.spritecollide(self, slideables, False)

        if player_hits or slideable_hits:
            # just moving the sprite down creates the illusion of a pushed button
            self.rect.topleft = (self.original_x, self.original_y + 2)
            self.pushed = True
            # This handles it, s.t. the Door Open Sound is only played once
            if self.pushed and not self.cont_push:
                self.door_open_sound.play()
                self.cont_push = True

        # restoring of original position and playing the door closing sound
        else:
            self.rect.topleft = (self.original_x, self.original_y)
            self.pushed = False

            if not self.pushed:
                if self.cont_push:
                    self.door_close_sound.play()
                    self.cont_push = False


class DoorTile(pygame.sprite.Sprite):
    def __init__(self, surface, position, door_id, connected_buttons):
        super().__init__()
        self.original_surface = surface
        self.surf = pygame.Surface((16,16), pygame.SRCALPHA)
        self.surf.blit(self.original_surface, dest=(0,0))
        self.rect = self.surf.get_rect(topleft=position)
        self.open = False
        self.door_id = door_id
        self.touched = False
        self.connected_buttons = [button["id"] for button in connected_buttons]
        self.open_state = -1
        self.open_timer = 0

    def update(self):
        # if there is any button, that is connected to us and pushed, we open
        if any([button.id in self.connected_buttons and button.pushed for button in sprites.buttons]):
            self.open = True

            # this is the animation system. Basically, we just adjust the transparency of our surface step by step
            # until we are invisible.
            if 0 <= self.open_state <= 5:
                self.open_timer += 1
                if self.open_timer % 10 == 0:
                    self.surf.set_alpha(int(255 - 255 * 0.2 * self.open_state))
                    self.open_state += 1
            elif self.open_state < 0:
                self.open_timer = 0
                self.open_state = 0
            else:
                self.open_timer = 0
                self.open_state = 6
                self.surf.set_alpha(0)
        else:
            self.open = False
            # Animation System as above. We decrease our transparency until we are fully visible again.
            if self.open_state >= 0 :
                self.open_timer += 1
                if self.open_timer % 10 == 0:
                    self.surf.set_alpha(int(255 - 255 * 0.2 * self.open_state))
                    self.open_state -= 1
            else:
                self.open_timer = 0
                self.open_state = -1
                self.surf.set_alpha(255)

class BackgroundTile:
    def __init__(self, surface, position):
        self.surf = surface
        self.pos = position


class StartTile(pygame.sprite.Sprite):
    def __init__(self, surface, position):
        super().__init__()
        self.surf = surface
        self.pos = position


class CoinTile(pygame.sprite.Sprite):
    def __init__(self, surface, position):
        super().__init__()
        self.surf = surface
        self.pos = position
        self.rect = self.surf.get_rect(topleft=position)
        self.animation_frames = []
        self.cur_animation_index = 0
        self.fps_counter = random.randint(0, 9)
        self.collected = False
        self.coin_sound = pygame.mixer.Sound("assets/sounds/coin_pickup.wav")

    def update(self):
        # Animation System
        self.cur_animation_index = self.cur_animation_index % len(self.animation_frames)

        self.fps_counter += 1
        if self.fps_counter % 10 == 0:
            self.fps_counter = 0
            self.cur_animation_index = (self.cur_animation_index + 1) % len(self.animation_frames)
        self.surf = self.animation_frames[self.cur_animation_index]

        hits = pygame.sprite.spritecollide(self, players, False)

        # If a player hits us, we are collected and thus increase the players score. Furthermore, we remove
        # ourselves from all sprite groups with self.kill and are thus not rendered anymore
        if hits and isinstance(hits[0], Player):
            self.collected = True
            hits[0].score += 1
            self.coin_sound.play()
            self.kill()


class BoneTile(CoinTile):
    def __init__(self, surface, position):
        super().__init__(surface=surface, position=position)
        self.bone_chant_channel = pygame.mixer.Channel(7)
        self.bone_chant = pygame.mixer.Sound("assets/sounds/bone_chant.mp3")

    def update(self):
        # Exactly the same as in the Coin, but a different sound is played and a different counter is updated

        self.cur_animation_index = self.cur_animation_index % len(self.animation_frames)

        self.fps_counter += 1
        if self.fps_counter % 10 == 0:
            self.fps_counter = 0
            self.cur_animation_index = (self.cur_animation_index + 1) % len(self.animation_frames)
        self.surf = self.animation_frames[self.cur_animation_index]

        hits = pygame.sprite.spritecollide(self, players, False)
        if hits and isinstance(hits[0], Player):
            self.collected = True
            hits[0].bone_score += 1
            self.bone_chant_channel.play(self.bone_chant)
            self.kill()


class FinishTile(pygame.sprite.Sprite):
    def __init__(self, surface, position):
        super().__init__()
        self.surf = surface
        self.rect = self.surf.get_rect(topleft=position)
        self.completed = False

    def update(self):
        hits = pygame.sprite.spritecollide(self, players, False)
        # Once both players collide with us, the level is completed)
        if len(hits) == 2:
            self.completed = True
        else:
            self.completed = False


class DeadTile(pygame.sprite.Sprite):
    def __init__(self, surface, position):
        super().__init__()
        self.surf = surface
        self.rect = pygame.Rect(position, (10, 12))
        self.rect.topleft = position + vec((16 - self.rect.width) // 2, 16 - self.rect.height)
        self.animation_frames = None
        self.animation_counter = 0
        self.cur_animation_index = 0
        self.killing = True

    def update(self):
        # Animation System
        if self.animation_frames:
            self.animation_counter += 1
            self.cur_animation_index = self.cur_animation_index % len(self.animation_frames)

            if self.animation_counter % 10 == 0:
                self.animation_counter = 0
                self.cur_animation_index = (self.cur_animation_index + 1) % len(self.animation_frames)
            self.surf = self.animation_frames[self.cur_animation_index]

        # If we collide with a slideable, we are not killing anymore. This enables the player to build bridges above
        # spikes. If this would not be here, because of too fast falling, the player collides with the box and the
        # dead tile (and is thus killed) if he jumps onto a bridge.
        slideables_collisions = pygame.sprite.spritecollide(self, slideables, False)
        if slideables_collisions and any(
                [slideable.rect.bottom - 1 == self.rect.bottom and abs(self.rect.left - slideable.rect.left <= 3) for slideable
                 in slideables_collisions]):
            self.killing = False
        else:
            self.killing = True
